package dev.fyloz.trial.colorrecipesexplorer.service

import dev.fyloz.trial.colorrecipesexplorer.exception.LowQuantitiesException
import dev.fyloz.trial.colorrecipesexplorer.exception.LowQuantityException
import dev.fyloz.trial.colorrecipesexplorer.model.MaterialQuantityDto
import dev.fyloz.trial.colorrecipesexplorer.model.materialQuantityDto
import dev.fyloz.trial.colorrecipesexplorer.service.utils.mapMayThrow
import org.springframework.stereotype.Service
import javax.transaction.Transactional

interface InventoryService {
    /** Adds each given [MaterialQuantityDto] to the inventory and returns the updated quantities. */
    fun add(materialQuantities: Collection<MaterialQuantityDto>): Collection<MaterialQuantityDto>

    /** Adds a given quantity to the given [Material]'s inventory quantity according to the given [materialQuantity] and returns the updated quantity. */
    fun add(materialQuantity: MaterialQuantityDto): Float

    /** Deducts the inventory quantity of each given [MaterialQuantityDto] and returns the updated quantities. */
    fun deduct(materialQuantities: Collection<MaterialQuantityDto>): Collection<MaterialQuantityDto>

    /** Deducts the inventory quantity of a given [Material] by a given quantity according to the given [materialQuantity] and returns the updated quantity. */
    fun deduct(materialQuantity: MaterialQuantityDto): Float
}

@Service
class InventoryServiceImpl(
    private val materialService: MaterialService
) : InventoryService {
    @Transactional
    override fun add(materialQuantities: Collection<MaterialQuantityDto>) =
        materialQuantities.map {
            materialQuantityDto(materialId = it.material, quantity = add(it))
        }

    override fun add(materialQuantity: MaterialQuantityDto) =
        materialService.updateQuantity(
            materialService.getById(materialQuantity.material),
            materialQuantity.quantity
        )

    @Transactional
    override fun deduct(materialQuantities: Collection<MaterialQuantityDto>): Collection<MaterialQuantityDto> {
        val thrown = mutableListOf<MaterialQuantityDto>()
        val updatedQuantities =
            materialQuantities.mapMayThrow<MaterialQuantityDto, MaterialQuantityDto, LowQuantityException>(
                { thrown.add(it.materialQuantity) }
            ) {
                materialQuantityDto(materialId = it.material, quantity = deduct(it))
            }

        if (thrown.isNotEmpty()) {
            throw LowQuantitiesException(thrown)
        }
        return updatedQuantities
    }

    override fun deduct(materialQuantity: MaterialQuantityDto): Float =
        with(materialService.getById(materialQuantity.material)) {
            if (this.inventoryQuantity >= materialQuantity.quantity) {
                materialService.updateQuantity(this, -materialQuantity.quantity)
            } else {
                throw LowQuantityException(materialQuantity)
            }
        }
}
