package dev.fyloz.trial.colorrecipesexplorer.service

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.trial.colorrecipesexplorer.model.*
import dev.fyloz.trial.colorrecipesexplorer.repository.MixRepository
import org.junit.jupiter.api.Test
import kotlin.test.assertEquals
import kotlin.test.assertTrue

class MixServiceTest : AbstractExternalModelServiceTest<Mix, MixSaveDto, MixUpdateDto, MixService, MixRepository>() {
    override val repository: MixRepository = mock()
    private val recipeService: RecipeService = mock()
    private val materialTypeService: MaterialTypeService = mock()
    private val mixMaterialService: MixMaterialService = mock()
    private val mixTypeService: MixTypeService = mock()
    override val service: MixService =
        spy(MixServiceImpl(repository, recipeService, materialTypeService, mixMaterialService, mixTypeService))

    override val entity: Mix = mix(id = 0L, location = "location")
    override val anotherEntity: Mix = mix(id = 1L)
    override val entitySaveDto: MixSaveDto = spy(mixSaveDto(mixMaterials = mapOf(1L to 1000f)))
    override val entityUpdateDto: MixUpdateDto = spy(mixUpdateDto(id = entity.id!!))

    // getAllByMixType()

    @Test
    fun `getAllByMixType() returns all mixes with the given mix type`() {
        val mixType = mixType(id = 0L)

        whenever(repository.findAllByMixType(mixType)).doReturn(entityList)

        val found = service.getAllByMixType(mixType)

        assertEquals(entityList, found)
    }

    // save()

    @Test
    override fun `save(dto) calls and returns save() with the created entity`() {
        val recipe = recipe(id = entitySaveDto.recipeId)
        val materialType = materialType(id = entitySaveDto.materialTypeId)
        val material = material(
            name = entitySaveDto.name,
            inventoryQuantity = Float.MIN_VALUE,
            isMixType = true,
            materialType = materialType
        )
        val mixType = mixType(name = entitySaveDto.name, material = material)
        val mix = mix(recipe = recipe, mixType = mixType)
        val mixWithId = mix(id = 0L, recipe = recipe, mixType = mixType)
        val mixMaterials = listOf(mixMaterial(mix = mixWithId, material = material(id = 1L), quantity = 1000f))

        whenever(recipeService.getById(recipe.id!!)).doReturn(recipe)
        whenever(materialTypeService.getById(materialType.id!!)).doReturn(materialType)
        whenever(mixMaterialService.createFromMap(mixWithId, entitySaveDto.mixMaterials!!)).doReturn(mixMaterials)
        whenever(
            mixTypeService.getOrCreateForNameAndMaterialType(
                mixType.name,
                mixType.material.materialType!!
            )
        ).doReturn(mixType)
        doReturn(true).whenever(service).existsById(mixWithId.id!!)
        doReturn(mixWithId).whenever(service).save(any<Mix>())
        doReturn(mixWithId).whenever(service).update(any<Mix>())

        val found = service.save(entitySaveDto)

        verify(service).save(argThat<Mix> { this.recipe == mix.recipe })
        verify(service).update(argThat<Mix> { this.id == mixWithId.id && this.recipe == mixWithId.recipe && this.mixMaterials == mixMaterials })
        verify(recipeService).addMix(recipe, mixWithId)

        // Verify if this method is called instead of the MixType's constructor, which does not check if the name is already taken by a material.
        verify(mixTypeService).getOrCreateForNameAndMaterialType(mixType.name, mixType.material.materialType!!)

        assertEquals(mixWithId, found)
    }

    // update()

    private fun mixUpdateDtoTest(
        scope: MixUpdateDtoTestScope = MixUpdateDtoTestScope(),
        sharedMixType: Boolean = false,
        op: MixUpdateDtoTestScope.() -> Unit
    ) {
        with(scope) {
            doReturn(true).whenever(service).existsById(mix.id!!)
            doReturn(mix).whenever(service).getById(mix.id!!)
            doReturn(sharedMixType).whenever(service).mixTypeIsShared(mix.mixType)
            doAnswer { it.arguments[0] }.whenever(service).update(any<Mix>())

            if (mixUpdateDto.materialTypeId != null) {
                whenever(materialTypeService.getById(materialType.id!!)).doReturn(materialType)
            }

            op()
        }
    }

    private fun mixUpdateDtoMixTypeTest(sharedMixType: Boolean = false, op: MixUpdateDtoTestScope.() -> Unit) {
        with(MixUpdateDtoTestScope(mixUpdateDto = mixUpdateDto(id = 0L, name = "name", materialTypeId = 0L))) {
            mixUpdateDtoTest(this, sharedMixType, op)
        }
    }

    @Test
    override fun `update(dto) calls and returns update() with the created entity`() {
        val mixUpdateDto = spy(mixUpdateDto(id = 0L, name = null, materialTypeId = null))

        doReturn(entity).whenever(service).getById(mixUpdateDto.id)
        doReturn(entity).whenever(service).update(entity)

        val found = service.update(mixUpdateDto)

        verify(service).update(entity)

        assertEquals(entity, found)
    }

    @Test
    fun `update(dto) calls MixTypeService saveForNameAndMaterialType() when mix type is shared`() {
        mixUpdateDtoMixTypeTest(sharedMixType = true) {
            whenever(mixTypeService.saveForNameAndMaterialType(mixUpdateDto.name!!, materialType))
                .doReturn(newMixType)

            val found = service.update(mixUpdateDto)

            verify(mixTypeService).saveForNameAndMaterialType(mixUpdateDto.name!!, materialType)

            assertEquals(newMixType, found.mixType)
        }
    }

    @Test
    fun `update(dto) calls MixTypeService updateForNameAndMaterialType() when mix type is not shared`() {
        mixUpdateDtoMixTypeTest {
            whenever(mixTypeService.updateForNameAndMaterialType(mixType, mixUpdateDto.name!!, materialType))
                .doReturn(newMixType)

            val found = service.update(mixUpdateDto)

            verify(mixTypeService).updateForNameAndMaterialType(mixType, mixUpdateDto.name!!, materialType)

            assertEquals(newMixType, found.mixType)
        }
    }

    @Test
    fun `update(dto) update, create and delete mix materials according to the given mix materials map`() {
        mixUpdateDtoTest {
            // Pairs exists, impairs don't
            val materials = listOf(
                material(id = 1L),
                material(id = 2L),
                material(id = 3L),
                material(id = 4L),
                material(id = 5L),
                material(id = 6L),
            )
            val toDelete = mixMaterial(material = material(id = 7L), quantity = 7000f)
            val mixMaterialsMap = mapOf(*materials.map { it.id!! to it.id!! * 1000f }.toTypedArray())
            val allMixMaterials: Collection<MixMaterial> = materials
                .map { mixMaterial(mix = mix, material = it, quantity = mixMaterialsMap[it.id]!!) }
            val existingMixMaterials = allMixMaterials.filter { it.material.id!! % 2 == 0L }.toMutableList()
            existingMixMaterials += toDelete

            mix.mixMaterials = existingMixMaterials
            (mixUpdateDto.mixMaterials as MutableMap<Long, Float>).putAll(mixMaterialsMap)

            doAnswer { allMixMaterials.first { mixMaterial -> mixMaterial.material.id == (it.arguments[0] as MixMaterial).material.id } }
                .whenever(mixMaterialService).updateQuantity(any(), any())
            doAnswer { allMixMaterials.first { mixMaterial -> mixMaterial.material.id == it.arguments[1] } }
                .whenever(mixMaterialService).create(eq(mix), any(), any())

            val found: Mix = service.update(mixUpdateDto)

            assertTrue { found.mixMaterials.containsAll(allMixMaterials) }

            verify(mixMaterialService, times(3)).updateQuantity(argThat { material.id!! % 2 == 0L }, any())
            verify(mixMaterialService, times(3)).create(eq(mix), any(), any())
            verify(mixMaterialService).delete(toDelete)
        }
    }

    // updateLocation()

    @Test
    fun `updateLocation() calls update() with the given mix with the given location`() {
        val newLocation = "new location"
        val expected = entity.apply { location = newLocation }
        doReturn(expected).whenever(service).update(expected)

        service.updateLocation(entity, newLocation)

        verify(service).update(expected)
    }

    // delete()

    override fun `deleteById() deletes the entity with the given id in the repository`() {
        super.`deleteById() deletes the entity with the given id in the repository`()
    }
}

data class MixUpdateDtoTestScope(
    val mixType: MixType = mixType(name = "mix type"),
    val newMixType: MixType = mixType(name = "another mix type"),
    val materialType: MaterialType = materialType(id = 0L),
    val mix: Mix = mix(id = 0L, mixType = mixType),
    val mixUpdateDto: MixUpdateDto = spy(
        mixUpdateDto(
            id = 0L,
            name = null,
            materialTypeId = null,
            mixMaterials = mutableMapOf()
        )
    )
)
